from __future__ import annotations

from agstoolbox.core.ags.ags_editor import LocalAgsEditor
from agstoolbox.core.ags.game_project import GameProject
from agstoolbox.core.utils.file import get_dir
from agstoolbox.wdgts_utils.file_explorer import open_folder


def ags_editor_folder_in_explorer(editor: LocalAgsEditor):
    open_folder(get_dir(editor.path))


def ags_project_folder_in_explorer(game_project: GameProject):
    open_folder(get_dir(game_project.path))



from __future__ import annotations  # for python 3.8
import os.path
from typing import Callable

from agstoolbox.core.gh.release import Release
from agstoolbox.core.settings.settings import ConstSettings
from agstoolbox.core.utils.downloader import download_from_url
from agstoolbox.core.utils.file import remove_dir_contents, mkdirp


def get_cache_dir(release: Release) -> str:
    return os.path.join(ConstSettings().cache_dir, "editor", release.version.as_str)


def get_zip_archive_cache_path(release: Release):
    cache_dir = get_cache_dir(release)
    return os.path.join(cache_dir, release.archive_name)


def download_release_to_cache(release: Release,
                              progress_update: Callable[[float, int, int, int], None] = None):
    r_url = release.archive_url
    filepath_in_cache = get_zip_archive_cache_path(release)
    remove_dir_contents(filepath_in_cache)
    cache_dir = get_cache_dir(release)
    mkdirp(cache_dir)
    download_from_url(url=r_url,
                      save_path=filepath_in_cache,
                      progress_update=progress_update)



from PyQt6 import QtWidgets, QtCore, QtGui
from PyQt6.QtWidgets import QMainWindow
from PyQt6.QtCore import QSize

from agstoolbox.core.settings.settings import ConstSettings
from agstoolbox.at_icons import icon_refresh, icon_settings, main_icon
from agstoolbox.panels.at_settings_dialog import SettingsDialog
from agstoolbox.wdgts.at_tree_projects_wdgt import ProjectsTree
from agstoolbox.wdgts.at_tree_tools_wdgt import ToolsTree


class MainWindow(QMainWindow):
    proj_update_task = None
    tool_update_task = None

    def __init__(self):
        QMainWindow.__init__(self)

        margin = 4

        self.setWindowTitle("AGS Toolbox")
        self.setObjectName("AgsToolbox")

        self.setWindowIcon(main_icon())

        self.setMinimumSize(QSize(ConstSettings().DEFAULT_MAIN_PANEL_WIDTH,
                                  ConstSettings().DEFAULT_MAIN_PANEL_HEIGHT))
        self.resize(QSize(ConstSettings().DEFAULT_MAIN_PANEL_WIDTH,
                          ConstSettings().DEFAULT_MAIN_PANEL_HEIGHT))
        self.setToolButtonStyle(QtCore.Qt.ToolButtonStyle.ToolButtonIconOnly)
        self.setUnifiedTitleAndToolBarOnMac(False)

        self.centralwidget = QtWidgets.QWidget(self)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout.setContentsMargins(margin, margin, margin, margin)
        self.verticalLayout.setObjectName("verticalLayout")
        self.tabWidget = QtWidgets.QTabWidget(self.centralwidget)
        self.tabWidget.setObjectName("tabWidget")

        # Tools
        self.tabTools = QtWidgets.QTabBar(self)
        self.tabTools.setObjectName("tabTools")
        self.tabTools.setDrawBase(False)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.tabTools)
        self.verticalLayout_2.setContentsMargins(margin, margin, margin, margin)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.treeTools = ToolsTree(parent=self)
        self.verticalLayout_2.addWidget(self.treeTools)
        self.tabWidget.addTab(self.tabTools, "")

        # Projects
        self.tabProjects = QtWidgets.QTabBar(self)
        self.tabProjects.setObjectName("tabProjects")
        self.tabProjects.setDrawBase(False)
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.tabProjects)
        self.verticalLayout_3.setContentsMargins(margin, margin, margin, margin)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.treeProjects = ProjectsTree(parent=self, toolsTree=self.treeTools)
        self.verticalLayout_3.addWidget(self.treeProjects)
        self.tabWidget.addTab(self.tabProjects, "")

        # back to main window things
        self.verticalLayout.addWidget(self.tabWidget)
        self.setCentralWidget(self.centralwidget)
        self.statusbar = QtWidgets.QStatusBar(self)
        self.statusbar.setObjectName("statusbar")
        self.setStatusBar(self.statusbar)

        self.progressBar = QtWidgets.QProgressBar()
        self.statusbar.addPermanentWidget(self.progressBar)
        self.progressBar.hide()

        # toolbar
        self.toolBar = QtWidgets.QToolBar(self)
        self.toolBar.setLayoutDirection(QtCore.Qt.LayoutDirection.RightToLeft)
        self.toolBar.setMovable(False)
        self.toolBar.setObjectName("toolBar")

        self.tabWidget.setCornerWidget(self.toolBar)
        # self.addToolBar(QtCore.Qt.ToolBarArea.TopToolBarArea, self.toolBar)

        self.actionSettings = QtGui.QAction(self)
        self.actionSettings.setIcon(icon_settings())
        self.actionSettings.setObjectName("actionSettings")
        self.actionSettings.triggered.connect(self.settings_clicked)

        self.actionRefresh = QtGui.QAction(self)
        self.actionRefresh.setIcon(icon_refresh())
        self.actionRefresh.setObjectName("actionRefresh")
        self.actionRefresh.triggered.connect(self.refresh_clicked)

        self.toolBar.addAction(self.actionSettings)
        self.toolBar.addAction(self.actionRefresh)

        self.retranslateUi()
        self.tabWidget.setCurrentIndex(0)

        self.settings_panel = SettingsDialog(parent=self)
        QtCore.QMetaObject.connectSlotsByName(self)

    def retranslateUi(self):
        _translate = QtCore.QCoreApplication.translate
        self.setWindowTitle(_translate("AgsToolbox", "AGS ToolBox"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tabTools),
                                  _translate("AgsToolbox", "BlueCup Rack"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tabProjects),
                                  _translate("AgsToolbox", "Projects"))
        self.toolBar.setWindowTitle(_translate("AgsToolbox", "toolBar"))
        self.actionSettings.setText(_translate("AgsToolbox", "Settings"))
        self.actionSettings.setToolTip(_translate("AgsToolbox", "Configure AGS Toolbox"))
        self.actionRefresh.setText(_translate("AgsToolbox", "Refresh"))
        self.actionRefresh.setToolTip(_translate("AgsToolbox", "check for new things to add now"))

    def refresh_all(self):
        self.treeProjects.projects_schd_update()
        self.treeTools.tools_schd_update_downloads()
        self.treeTools.tools_schd_update_managed()
        self.treeTools.tools_schd_update_unmanaged()

    def refresh_clicked(self):
        self.refresh_all()

    def settings_clicked(self):
        self.settings_panel.show()



from __future__ import annotations

import errno
import glob
import os
import shutil
from os.path import realpath
from os.path import dirname
from os.path import join as join_paths
from pathlib import Path


def get_dir(filepath: str):
    return Path(realpath(dirname(filepath))).as_posix()


def get_file(filepath: str):
    return os.path.basename(Path(filepath).as_posix())


def get_absolute_path(filepath: str) -> str:
    return os.path.abspath(os.path.expanduser(os.path.expandvars(filepath)))


def dir_is_valid(folderpath: str) -> bool:
    res: bool = False
    try:
        res = folderpath is not None and \
              not "".__eq__(folderpath) and \
              Path(folderpath).exists() and \
              Path(folderpath).is_dir()
    except Exception as e:
        res = False
    finally:
        return res


def get_valid_dirs(directories: list[str]) -> list[str]:
    return [d for d in directories if dir_is_valid(d)]


def get_unique_list(lst: list[str]) -> list[str]:
    return list(dict.fromkeys(lst))


def get_unique_valid_dirs(directories: list[str]) -> list[str]:
    return get_valid_dirs(get_unique_list(directories))


# TODO: do this non-hacky way!
def get_file_if_exists(directory: str, file: str):
    path_test_1 = os.path.join(directory, file.upper())
    path_test_2 = os.path.join(directory, file.lower())
    ftks = file.rsplit(".", 2)
    name = ftks[0]
    ext = ""
    if len(ftks) > 1:
        ext = ftks[1]

    path_test_3 = os.path.join(directory, name.upper() + "." + ext.lower())
    path_test_4 = os.path.join(directory, name.lower() + "." + ext.upper())

    if Path(path_test_1).exists():
        return path_test_1
    if Path(path_test_2).exists():
        return path_test_2
    if Path(path_test_3).exists():
        return path_test_3
    if Path(path_test_4).exists():
        return path_test_4

    return None


def internal_run_fast_scandir(directory: str, filename: str):  # dir: str, ext: list
    sub_folders, files = [], []

    if Path(directory).is_dir() and Path(directory).exists():
        directory = Path(directory).as_posix()
    else:
        directory = get_dir(directory)

    os_scandir_res = []
    try:
        os_scandir_res = os.scandir(directory)
    except NotADirectoryError:
        os_scandir_res = []
    except PermissionError:
        os_scandir_res = []

    for f in os_scandir_res:
        if f.is_dir():
            sub_folders.append(f.path)
        if f.is_file():
            if f.name == filename:
                files.append(f.path)

    for s_dir in list(sub_folders):
        sf, f = internal_run_fast_scandir(s_dir, filename)
        sub_folders.extend(sf)
        files.extend(f)
    return sub_folders, files


def get_gp_candidates_in_dir_fastscandir(directory: str, filename: str) -> list[str]:
    # in Python 3.8, this is ~25% faster, but python 3.10 it's slower
    # unfortunately, we are using Python 3.8 for the time being
    _, files = internal_run_fast_scandir(directory, filename)
    return files


def get_gp_candidates_in_dir_glob(directory: str, filename: str) -> list[str]:
    pathname = directory + "/**/" + filename
    return glob.glob(pathname, recursive=True)


def get_gp_candidates_in_dir(directory: str, filename: str) -> list[str]:
    if not os.path.exists(directory):
        return []

    files = get_gp_candidates_in_dir_fastscandir(directory, filename)
    if not files:
        return []

    files = [Path(f).as_posix() for f in files]

    return files


def join_paths_as_posix(path_first: str, path_second: str) -> str:
    return Path(os.path.join(path_first, path_second)).as_posix()


def remove_dir_contents(target_dir: str):
    try:
        with os.scandir(target_dir) as entries:
            for entry in entries:
                if entry.is_dir() and not entry.is_symlink():
                    shutil.rmtree(entry.path)
                elif entry.is_file():
                    os.remove(entry.path)
    except (FileNotFoundError, IOError):
        # print("nothing to remove")
        pass


def _mk_dir_recursive(dir_path):
    if os.path.isdir(dir_path):
        return
    h, t = os.path.split(dir_path)  # head/tail
    if not os.path.isdir(h):
        _mk_dir_recursive(h)

    new_path = join_paths(h, t)
    if not os.path.isdir(new_path):
        try:
            os.mkdir(str(new_path))
        except OSError as exc:
            if exc.errno != errno.EEXIST:
                raise
            pass


def mkdirp(path_dir: str):
    _mk_dir_recursive(path_dir)
    # alternate implementation if needed
    # Path(path_dir).mkdir(parents=True, exist_ok=True)



from __future__ import annotations  # for python 3.8

from datetime import datetime
from operator import attrgetter

import requests

from agstoolbox.core.gh.release import Release
from agstoolbox.core.version.version import Version
from agstoolbox.core.version.version_utils import tag_to_version


def is_asset_archive(release_name: str, asset_name: str) -> bool:
    version = release_name.replace(" ", "").replace("v.", "")
    is_patch = asset_name.startswith("AGS-" + version + "-P")
    if is_patch:
        patch = asset_name.split(version + "-P")[1].split(".")[0]
        version += "-P" + patch

    archive_name = "AGS-" + version + ".zip"
    return asset_name == archive_name


def parse_releases(response_json) -> list[Release]:
    releases = [None] * len(response_json)
    count = 0

    for rel in response_json:
        rls = None
        found_asset = False

        for asset in rel['assets']:
            # check for either predictable or patch release archives
            if is_asset_archive(rel['name'], asset['name']):
                rls = Release()
                rls.archive_id = str(asset['id'])
                rls.archive_name = asset['name']
                rls.archive_url = asset['browser_download_url']
                rls.archive_size = int(asset['size'])
                found_asset = True
                break

        if not found_asset:
            for asset in rel['assets']:
                # check for either predictable or patch release archives
                if is_asset_archive(rel['tag_name'], asset['name']):
                    rls = Release()
                    rls.archive_id = str(asset['id'])
                    rls.archive_name = asset['name']
                    rls.archive_url = asset['browser_download_url']
                    rls.archive_size = asset['size']
                    found_asset = True
                    break

        if not found_asset:
            continue

        rls.text_details = rel['body']
        rls.name = rel['name']
        rls.id = str(rel['id'])
        rls.url = rel['url']
        rls.html_url = rel['html_url']

        tag = rel['tag_name']
        rls.tag = tag
        rls.version = tag_to_version(tag)

        rls.is_pre_release = rel['prerelease']
        rls.published_at = rel['published_at']
        tstamp = datetime.strptime(rel['published_at'], "%Y-%m-%dT%H:%M:%SZ")
        rls.published_at_timestamp = tstamp.timestamp()
        releases[count] = rls
        count += 1

    releases = list(filter(None, releases))

    releases.sort(key=attrgetter("published_at_timestamp"), reverse=True)

    return releases


def list_releases() -> list[Release]:
    response = requests.get(
        "https://api.github.com/repos/adventuregamestudio/ags/releases?per_page=100")
    response_json = response.json()
    return parse_releases(response_json)


def get_latest_release_family(releases: list[Release], family: str) -> Release | None:
    filtered_releases = [r for r in releases if r.version.family == family]
    filtered_releases.sort(key=attrgetter("version.as_int"), reverse=True)
    if len(filtered_releases) >= 1:
        return filtered_releases[0]
    else:
        return None


def get_release_version(releases: list[Release], version: Version) -> Release | None:
    filtered_releases = [r for r in releases if r.version.as_int == version.as_int]
    if len(filtered_releases) == 1:
        return filtered_releases[0]
    else:
        return None



from __future__ import annotations  # for python 3.8
from operator import attrgetter

from PyQt6.QtWidgets import QTreeWidget, QWidget, QFrame

from agstoolbox.at_tasks import do_update_projects
from agstoolbox.wdgts.at_tree_item_project import TreeItemProject
from agstoolbox.wdgts.at_tree_tools_wdgt import ToolsTree


class ProjectsTree(QTreeWidget):
    proj_update_task = None
    tools_tree: ToolsTree = None

    def __init__(self, parent: QWidget = None, toolsTree: ToolsTree = None):
        QTreeWidget.__init__(self, parent)

        self.tools_tree = toolsTree
        self.setHeaderHidden(True)
        self.setObjectName("treeProjects")
        self.setFrameStyle(QFrame.Shape.NoFrame)
        self.setRootIsDecorated(False)
        self.setIndentation(0)

    # AGS Projects stuff
    def projects_schd_update(self):
        if self.proj_update_task is not None:
            return

        self.proj_update_task = do_update_projects(self.projects_update, self.projects_update_ended)

    def projects_update(self):
        self.clear()
        projs = self.proj_update_task.proj_list

        projs.sort(key=attrgetter("last_modified"), reverse=True)

        for p in projs:
            itm = TreeItemProject(ags_game_project=p)
            self.addTopLevelItem(itm)
            itm.updateInTree()

        self.projects_update_ended()

    def projects_update_ended(self):
        self.proj_update_task = None



from __future__ import annotations  # for python 3.8

import os
from subprocess import Popen, TimeoutExpired

from agstoolbox.core.utils.file import get_file, get_dir


def run_exe_params(exe_path: str, block: bool = False, params: list[str] = []):
    exe_file = get_file(exe_path)
    working_dir = get_dir(exe_path)

    p_params = [exe_file]
    p_params.extend(params)

    cwd = os.getcwd()
    os.chdir(working_dir)
    print('Popen: cwd=' + working_dir + ', ' + ' '.join(p_params))
    proc = Popen(p_params, cwd=working_dir)
    count = 0
    if block:
        while count < 1000:
            try:
                proc.wait(10)
            except TimeoutExpired:
                pass
            count += 1

        if count == 1000:
            proc.terminate()

    os.chdir(cwd)



from __future__ import annotations  # for python 3.8
from pathlib import Path

from PyQt6 import QtCore, QtWidgets
from PyQt6.QtWidgets import QWidget, QFileDialog, QStyle, QFrame

from agstoolbox.core.settings.settings import ConstSettings
from agstoolbox.core.utils.file import dir_is_valid


class DirListWidget(QWidget):
    default_dirs_value = None

    def is_selection_valid(self) -> bool:
        if self.list.count() == 0:
            return False

        itm = self.getSelectedItem()
        if itm is None:
            return False
        row = self.list.row(itm)
        if row >= self.list.count():
            return False
        return True

    def appendDir(self, d):
        itm = QtWidgets.QListWidgetItem(d, parent=self.list)
        self.list.addItem(itm)

    def setDirs(self, dirs: list[str]):
        self.list.clear()
        for d in dirs:
            self.appendDir(d)

    def getSelectedItem(self) -> QtWidgets.QListWidgetItem:
        sel = self.list.selectedItems()

        i = self.list.count() - 1
        itm = self.list.item(i)
        if sel is not None and len(sel) > 0:
            itm = sel[0]

        return itm

    def getDirs(self) -> list[str]:
        dirs = []
        for i in range(self.list.count()):
            itm = self.list.item(i)
            d = itm.text()
            dirs.append(d)

        return dirs

    def btn_new_clicked(self):
        dir_path = QFileDialog.getExistingDirectory(
            self, 'Select Folder',
            options=QFileDialog.Option.ShowDirsOnly,
            directory=ConstSettings().user_docs
        )

        if dir_path is None or len(dir_path) <= 1:
            return

        if not dir_is_valid(dir_path):
            return

        dir_path = str(Path(str(dir_path)).as_posix())

        self.appendDir(dir_path)

    def btn_edit_clicked(self):
        if self.list.count() == 0:
            return

        itm = self.getSelectedItem()
        dir_path = itm.text()
        if not dir_is_valid(dir_path):
            dir_path = ConstSettings().user_docs

        dir_path = QFileDialog.getExistingDirectory(
            self, 'Select Folder',
            options=QFileDialog.Option.ShowDirsOnly,
            directory=dir_path)

        if not dir_is_valid(dir_path):
            return

        dir_path = str(Path(str(dir_path)).as_posix())
        itm.setText(dir_path)

    def btn_del_clicked(self):
        if not self.is_selection_valid():
            return

        itm = self.getSelectedItem()
        row = self.list.row(itm)
        self.list.takeItem(row)
        del itm

    def btn_defaults_clicked(self):
        self.setDirs(self.default_dirs_value)

    def btn_move_up_clicked(self):
        if not self.is_selection_valid():
            return
        if self.list.count() <= 1:
            return

        itm = self.getSelectedItem()
        row = self.list.row(itm)
        if row > 0:
            self.list.takeItem(row)
            target_row = row - 1
            self.list.insertItem(target_row, itm)
            self.list.setCurrentRow(target_row)

    def btn_move_down_clicked(self):
        if not self.is_selection_valid():
            return
        if self.list.count() <= 1:
            return

        itm = self.getSelectedItem()
        row = self.list.row(itm)
        if row < self.list.count() - 1:
            self.list.takeItem(row)
            target_row = row + 1
            self.list.insertItem(target_row, itm)
            self.list.setCurrentRow(target_row)

    def __init__(self, dirs: list[str], default_dirs: list[str], parent: QWidget = None):
        QWidget.__init__(self, parent)

        self.setObjectName("dir_list_widget")

        self.default_dirs_value = default_dirs

        self.list = QtWidgets.QListWidget(self)
        self.list.setObjectName("list")
        self.list.setFrameStyle(QFrame.Shape.NoFrame)
        self.list.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.SingleSelection)
        self.setDirs(dirs)

        self.push_button_new = QtWidgets.QPushButton(self)
        self.push_button_new.setObjectName("push_button_new")
        self.push_button_edit = QtWidgets.QPushButton(self)
        self.push_button_edit.setObjectName("push_button_edit")
        self.push_button_del = QtWidgets.QPushButton(self)
        self.push_button_del.setObjectName("push_button_del")

        self.push_button_move_up = QtWidgets.QPushButton(self)
        self.push_button_move_up.setObjectName("push_button_move_up")
        icon = self.style().standardIcon(QStyle.StandardPixmap.SP_ArrowUp)
        self.push_button_move_up.setIcon(icon)

        self.push_button_move_down = QtWidgets.QPushButton(self)
        self.push_button_move_down.setObjectName("push_button_move_down")
        icon = self.style().standardIcon(QStyle.StandardPixmap.SP_ArrowDown)
        self.push_button_move_down.setIcon(icon)

        self.push_button_defaults = QtWidgets.QPushButton(self)
        self.push_button_defaults.setObjectName("push_button_defaults")

        # Do Layout
        spacer_item_fixed = QtWidgets.QSpacerItem(
            20, 16, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Fixed)
        spacer_item_expanding = QtWidgets.QSpacerItem(
            20, 16, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Expanding)
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.verticalLayout.addWidget(self.push_button_new)
        self.verticalLayout.addWidget(self.push_button_edit)
        self.verticalLayout.addWidget(self.push_button_del)
        self.verticalLayout.addItem(spacer_item_fixed)

        self.h_move_layout = QtWidgets.QHBoxLayout()
        self.h_move_layout.setObjectName("h_move_layout")
        self.h_move_layout.addWidget(self.push_button_move_up)
        self.h_move_layout.addWidget(self.push_button_move_down)

        self.verticalLayout.addLayout(self.h_move_layout)
        self.verticalLayout.addWidget(self.push_button_defaults)
        self.verticalLayout.addItem(spacer_item_expanding)

        self.horizontalLayout = QtWidgets.QHBoxLayout(self)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.horizontalLayout.addWidget(self.list)
        self.horizontalLayout.addLayout(self.verticalLayout)
        self.setLayout(self.horizontalLayout)

        self.retranslateUi()

        # Connect events
        self.push_button_new.clicked.connect(self.btn_new_clicked)
        self.push_button_del.clicked.connect(self.btn_del_clicked)
        self.push_button_edit.clicked.connect(self.btn_edit_clicked)
        self.push_button_defaults.clicked.connect(self.btn_defaults_clicked)
        self.push_button_move_up.clicked.connect(self.btn_move_up_clicked)
        self.push_button_move_down.clicked.connect(self.btn_move_down_clicked)

    def retranslateUi(self):
        _translate = QtCore.QCoreApplication.translate

        parent = "dir_list_widget"
        if self.parent() is not None:
            parent = self.parent().objectName()

        self.push_button_new.setText(_translate(parent, "New"))
        self.push_button_edit.setText(_translate(parent, "Edit"))
        self.push_button_del.setText(_translate(parent, "Delete"))
        self.push_button_defaults.setText(_translate(parent, "Set Defaults"))





from __future__ import annotations  # for python 3.8


def get_str_list_from_dict(data: dict, key: str) -> list[str]:
    content = None
    try:
        content = data[key]
        if type(content) == type(str()):  # noqa: E721
            content = [content]
    except KeyError:
        content = None
    finally:
        return content


def get_str_from_dict(data: dict, key: str) -> str:
    content = None
    try:
        content = data[key]
    except KeyError:
        content = None
    finally:
        return content


def get_bool_from_dict(data: dict, key: str) -> bool:
    content = None
    try:
        content = data[key]
    except KeyError:
        content = None
    finally:
        if content is not None:
            content = content and True

        return content




from __future__ import annotations  # for python 3.8

from agstoolbox.core.cmdline.progress import ProgressBar
from agstoolbox.core.gh.download_release import download_release_to_cache
from agstoolbox.core.gh.release import Release


def cmdline_download_release_to_cache(release: Release, quiet: bool):
    max_size: int = release.archive_size

    bar = None
    if not quiet:
        bar = ProgressBar(' Downloading... ', ' ' + release.archive_name + ' ', max_size, 'B')

    def progress_updates(completed_percent: float,
                         c_size: int,
                         acc_c_size: int,
                         total_size_in_bytes: int):
        if bar is not None:
            bar.update(int(acc_c_size/8), total_size_in_bytes)

    download_release_to_cache(release, progress_updates)

    if bar is not None:
        bar.finish()



from __future__ import annotations  # for python 3.8

from pathlib import Path

from platformdirs import user_cache_dir, user_data_dir, user_log_dir, user_documents_dir


def get_user_cache_dir(app_name: str, app_author: str) -> str:
    return Path(user_cache_dir(app_name, app_author)).as_posix()


def get_user_data_dir(app_name: str, app_author: str) -> str:
    return Path(user_data_dir(app_name, app_author)).as_posix()


def get_user_log_dir(app_name: str, app_author: str) -> str:
    return Path(user_log_dir(app_name, app_author)).as_posix()


def get_user_documents_dir() -> str:
    return Path(user_documents_dir()).as_posix()



from __future__ import annotations  # for python 3.8
from datetime import datetime, timedelta


def internal_s_ago(t: datetime, diff: timedelta):
    """
        Calculate a '3 hours ago' type string from a python datetime,
        if time less than 5 days, otherwise, return date as day/month/year
    """
    units = {
        'days': lambda diff_t: diff_t.days,
        'hours': lambda diff_t: int(diff_t.seconds / 3600),
        'minutes': lambda diff_t: int(diff_t.seconds % 3600 / 60),
    }

    if diff.days >= 5:
        return t.strftime("%d/%m/%Y")

    for unit in units:
        dur = units[unit](diff)  # Run the lambda function to get a duration
        if dur > 0:
            # if duration is 1 ('1 day' vs '2 days')
            unit = unit[:-dur] if dur == 1 else unit  # De-pluralize
            return '%s %s ago' % (dur, unit)
    return 'just now'


def s_ago(ft: float) -> str:
    t = datetime.utcfromtimestamp(ft)
    diff = datetime.utcnow() - t

    return internal_s_ago(t, diff)



from __future__ import annotations

import json
import os
from pathlib import Path
from platform import platform

from agstoolbox.core.settings.settings_data import SettingsData
from agstoolbox.core.utils.basics import get_str_list_from_dict, get_str_from_dict, \
    get_bool_from_dict


def load_settings_data_from_json_string(json_string: str) -> SettingsData:
    sd = SettingsData()

    if json_string is None:
        return sd

    try:
        data = json.loads(json_string)
    except ValueError:
        return sd

    if data is None:
        return sd

    sd.manually_installed_editors_search_dirs = get_str_list_from_dict(
        data, 'manually_installed_editors_search_dirs')
    sd.project_search_dirs = get_str_list_from_dict(
        data, 'project_search_dirs')
    sd.tools_install_dir = get_str_from_dict(
        data, 'tools_install_dir')
    sd.run_when_os_starts = get_bool_from_dict(
        data, 'run_when_os_starts')
    return sd


def save_settings_data_to_json_string(settings_data: SettingsData) -> str:
    data = {
        "tools_install_dir": settings_data.tools_install_dir,
        "project_search_dirs": settings_data.project_search_dirs,
        "manually_installed_editors_search_dirs":
            settings_data.manually_installed_editors_search_dirs,
        "run_when_os_starts": settings_data.run_when_os_starts
    }

    data = {k: v for k, v in data.items() if v is not None}

    json_string = json.dumps(data, indent=4, sort_keys=True) + "\n"
    return json_string


def win_get_default_editor_search_dirs():
    if not platform().lower().startswith('win'):
        return []

    versions = ['3.4.3', '3.5.0', '3.5.1', '3.6.0', '3.99.99', '3.99.100', '4.0.0']
    ret = []
    dirs = []
    p_files1 = os.environ["ProgramFiles"]
    p_files2 = os.environ["ProgramFiles(x86)"]
    p_files3 = os.environ["ProgramW6432"]
    if len(p_files1) > 1:
        dirs.append(p_files1)
    if len(p_files2) > 1:
        dirs.append(p_files2)
    if len(p_files3) > 1:
        dirs.append(p_files3)

    for v in versions:
        for d in dirs:
            ags_ed_d = os.path.join(d, 'Adventure Game Studio ' + v)
            if Path(ags_ed_d).exists():
                ret.append(ags_ed_d)

    ret = list(dict.fromkeys(ret))

    return ret



from __future__ import annotations  # for python 3.8

from operator import attrgetter
from pathlib import Path
from sys import exit
import argparse
from typing import Callable

import shtab  # for completion magic

from agstoolbox import __title__, __version__, __copyright__, __license__
from agstoolbox.core.ags.ags_editor import LocalAgsEditor
from agstoolbox.core.ags.ags_local_run import ags_editor_load_project, start_ags_editor, \
    ags_editor_build_project
from agstoolbox.core.ags.game_project import GameProject
from agstoolbox.core.ags.get_game_projects import list_game_projects_in_dir, \
    list_game_projects_in_dir_list, get_unique_game_project_in_path
from agstoolbox.core.ags.get_local_ags_editors import list_probable_ags_editors_in_dir, \
    list_ags_editors_in_dir_list
from agstoolbox.core.cmdline.cmdline_download import cmdline_download_release_to_cache
from agstoolbox.core.gh.install_release import is_install_dir_busy, install_release_from_cache
from agstoolbox.core.gh.list_releases import list_releases, get_latest_release_family, \
    get_release_version
from agstoolbox.core.gh.release import Release
from agstoolbox.core.settings.settings import Settings
from agstoolbox.core.version.version import Version
from agstoolbox.core.version.version_utils import version_str_to_version


def meta_cmd_project(args,
                     ags_editor_proj_command: Callable[
                         [LocalAgsEditor, GameProject, bool], None] = None):
    block: bool = not args.non_blocking
    prj_path: str = args.PROJECT_PATH

    if not Path(prj_path).exists():
        print('ERROR: Invalid project path')
        return

    game_project: GameProject | None = get_unique_game_project_in_path(prj_path)
    if game_project is None:
        print('ERROR: Invalid project path')
        return

    project_version: Version = game_project.ags_editor_version

    managed_dir: str = Settings().get_tools_install_dir()
    editors = list_probable_ags_editors_in_dir(managed_dir)

    for editor in editors:
        if editor.version.as_int == project_version.as_int:
            ags_editor_proj_command(editor, game_project, block)
            return

    unmanaged_dirs: list[str] = Settings().get_manually_installed_editors_search_dirs()
    editors = list_ags_editors_in_dir_list(unmanaged_dirs)

    for editor in editors:
        if editor.version.as_int == project_version.as_int:
            ags_editor_proj_command(editor, game_project, block)
            return

    print("ERROR: Failed to find exact match of AGS Editor, wanted " + project_version.as_str)


def at_cmd_list_projects(args):
    a_path: str | None = args.path

    if a_path is None:
        Settings().load()
        prj_dirs: list[str] = Settings().get_project_search_dirs()
        projects = list_game_projects_in_dir_list(prj_dirs)

    else:
        projects = list_game_projects_in_dir(a_path)

    for p in projects:
        print(p.name + ', ' + p.ags_editor_version.as_str + ', ' + p.path)
    pass


def at_cmd_list_editors(args):
    is_unmanaged: bool = args.unmanaged and True
    is_download: bool = args.download and True
    a_path: str | None = args.path

    releases: list[Release] = []
    editors: list[LocalAgsEditor] = []
    if a_path is None:
        Settings().load()
        if is_unmanaged:
            unmanaged_dirs: list[str] = Settings().get_manually_installed_editors_search_dirs()
            editors = list_ags_editors_in_dir_list(unmanaged_dirs)
        elif is_download:
            releases = list_releases()
        else:
            managed_dir: str = Settings().get_tools_install_dir()
            editors = list_probable_ags_editors_in_dir(managed_dir)

    else:
        editors = list_probable_ags_editors_in_dir(a_path)

    for e in editors:
        print(e.name + ', ' + e.version.as_str + ', ' + e.path)
    pass

    for r in releases:
        print('download' + ', ', r.version.as_str + ', ' + r.html_url)
    pass


def at_cmd_list(args):
    is_editor_list: bool = args.sub_list == 'editors'
    is_proj_list: bool = args.sub_list == 'projects'

    if is_proj_list:
        at_cmd_list_projects(args)
    elif is_editor_list:
        at_cmd_list_editors(args)
    else:
        print('ERROR: Invalid list command!')


def at_cmd_install(args):
    quiet: bool = True and args.quiet
    force: bool = True and args.force
    install_arg: str = args.VERSION_OR_PROJECT_PATH

    if install_arg is None or install_arg == "":
        print('ERROR: Editor Version specified is invalid!')
        return

    editor_version: Version | None = None
    if Path(install_arg).exists():
        game_project: GameProject | None = get_unique_game_project_in_path(install_arg)
        if game_project is not None:
            editor_version = game_project.ags_editor_version

    release_to_install: Release = Release()
    releases: list[Release] = list_releases()

    if editor_version is None or editor_version.as_int < 3000000000:
        # arg was really a version
        editor_version = version_str_to_version(install_arg)
        if editor_version.improv == "0" and editor_version.patch == "0":
            release_to_install = get_latest_release_family(releases, editor_version.family)

    ri = release_to_install
    if ri is None or ri.archive_url is None or len(ri.archive_url) <= 1:
        release_to_install = get_release_version(releases, editor_version)

    if release_to_install is None:
        print('ERROR: Editor Version specified is invalid!')
        return

    if is_install_dir_busy(release_to_install) and not force:
        print('ERROR: Editor Version already installed!')
        return

    print("Will install managed AGS Editor release " + release_to_install.version.as_str)

    cmdline_download_release_to_cache(release_to_install, quiet)

    print("Extracting...")
    install_release_from_cache(release_to_install)
    print("Installed release " + release_to_install.version.as_str)
    pass


def at_cmd_open_editor(args):
    editor_version: Version = version_str_to_version(args.EDITOR_VERSION)

    if editor_version is None or editor_version.as_int < 3000000000:
        print('ERROR: Invalid version!')
        return

    managed_dir: str = Settings().get_tools_install_dir()
    editors = list_probable_ags_editors_in_dir(managed_dir)

    for editor in editors:
        if editor.version.as_int == editor_version.as_int:
            start_ags_editor(editor)
            return

    unmanaged_dirs: list[str] = Settings().get_manually_installed_editors_search_dirs()
    un_editors = list_ags_editors_in_dir_list(unmanaged_dirs)

    for editor in un_editors:
        if editor.version.as_int == editor_version.as_int:
            start_ags_editor(editor)
            return

    print("WARN: Failed to find exact match of AGS Editor( " + editor_version.as_str +
          " ), will try to find a compatible one")

    filtered_editors = [ae for ae in editors if ae.version.family == editor_version.family]
    filtered_editors.sort(key=attrgetter("version.as_int"), reverse=True)
    if len(filtered_editors) >= 1:
        start_ags_editor(filtered_editors[0])
        return

    filtered_un_editors = [ae for ae in un_editors if ae.version.family == editor_version.family]
    filtered_un_editors.sort(key=attrgetter("version.as_int"), reverse=True)
    if len(filtered_un_editors) >= 1:
        start_ags_editor(filtered_un_editors[0])
        return

    print("ERROR: No compatible AGS Editor available, wanted " + editor_version.family)


def at_cmd_open_project(args):
    meta_cmd_project(args, ags_editor_load_project)
    return


def at_cmd_open(args):
    is_editor_open: bool = args.sub_open == 'editor'
    is_proj_open: bool = args.sub_open == 'project'

    if is_proj_open:
        at_cmd_open_project(args)
    elif is_editor_open:
        at_cmd_open_editor(args)
    else:
        print('ERROR: Invalid open command!')


def at_cmd_build(args):
    meta_cmd_project(args, ags_editor_build_project)
    return


def at_cmd_settings_show(args):
    show_what: str | None = args.sub_setting_param

    if show_what is None or show_what == 'all':
        print(Settings().dump())
    elif show_what == 'project_search_dirs':
        print("Project Search Dirs:", ', '.join(Settings().get_project_search_dirs()))
    elif show_what == 'tools_install_dir':
        print("Tools Install Dir:", Settings().get_tools_install_dir())
    else:
        print('ERROR: Invalid settings show command!')
        return


def at_cmd_settings_set(args):
    set_what: str | None = args.sub_setting_param

    if set_what == 'tools_install_dir':
        Settings().set_tools_install_dir(args.value)
    else:
        print('ERROR: Invalid settings set command!')
        return

    Settings().save()


def at_cmd_settings(args):
    is_show_settings: bool = args.sub_settings == 'show'
    is_set_settings: bool = args.sub_settings == 'set'

    if is_show_settings:
        at_cmd_settings_show(args)
        return
    elif is_set_settings:
        at_cmd_settings_set(args)
        return
    else:
        print('ERROR: Invalid settings command!')
        return


def cmdline(show_help_when_empty: bool, program_name: str):
    parser = argparse.ArgumentParser(
        prog=program_name,
        description=__title__ + ' is an application to help manage AGS Editor versions.',
        epilog=__copyright__ + ", " + __license__ + ".")

    shtab.add_argument_to(parser, ["-s"])  # magic!
    parser.add_argument(
        '-v', '--version', action='store_true', default=False, help='get software version.')
    subparsers = parser.add_subparsers(help='command')

    # list command
    p_l = subparsers.add_parser('list', help='lists things')
    p_l.set_defaults(func=at_cmd_list)
    p_ll = p_l.add_subparsers(title='sub_list', dest='sub_list')
    p_lle = p_ll.add_parser('editors', help='lists AGS Editors, by default only managed editors')
    p_lle.add_argument('-p', '--path', action='store', default=None, type=str,
                       help='look for AGS Editors in specific path, ignore settings'
                       ).complete = shtab.DIRECTORY
    p_lle.add_argument(
        '-u', '--unmanaged', action='store_true', default=False,
        help='search for unmanaged editors in settings')
    p_lle.add_argument(
        '-d', '--download', action='store_true', default=False,
        help='lists for editors available for download')
    p_llp = p_ll.add_parser('projects', help='lists AGS Projects')
    p_llp.add_argument('-p', '--path', action='store', default=None, type=str,
                       help='the path to look for list').complete = shtab.DIRECTORY

    # install command
    p_i = subparsers.add_parser('install', help='install tools')
    p_i.set_defaults(func=at_cmd_install)
    p_ii = p_i.add_subparsers(title='sub_install', dest='sub_install')
    p_iie = p_ii.add_parser('editor', help='install managed AGS Editor')
    p_iie.add_argument('VERSION_OR_PROJECT_PATH',
                       help='version to install or project to grab editor version')
    p_iie.add_argument('-f', '--force', action='store_true', default=None,
                       help='forces installation, overwrite if already exists')
    p_iie.add_argument('-q', '--quiet', action='store_true', default=False,
                       help='do not print download progress')

    # open command
    p_o = subparsers.add_parser('open', help='open an editor or project')
    p_o.set_defaults(func=at_cmd_open)
    p_oo = p_o.add_subparsers(title='sub_open', dest='sub_open')
    p_ooe = p_oo.add_parser('editor', help='open AGS Editor, by default only managed editors')
    p_ooe.add_argument('EDITOR_VERSION',
                       help='Editor version to open')
    p_oop = p_oo.add_parser('project', help='open AGS Project')
    p_oop.add_argument('PROJECT_PATH',
                       help='path to the project to be opened').complete = shtab.FILE
    p_oop.add_argument('-n', '--non-blocking', action='store_true', default=False,
                       help='do not wait for Editor to be closed to continue')

    # build command
    p_b = subparsers.add_parser('build', help='build an ags project')
    p_b.set_defaults(func=at_cmd_build)
    p_b.add_argument('PROJECT_PATH',
                     help='path to the project to be built').complete = shtab.FILE
    p_b.add_argument('-n', '--non-blocking', action='store_true', default=False,
                     help='do not wait for Editor to be closed to continue')

    # settings command
    p_s = subparsers.add_parser('settings', help='modify or show settings')
    p_s.set_defaults(func=at_cmd_settings)
    p_ss = p_s.add_subparsers(title='sub_settings', dest='sub_settings')

    p_ssh = p_ss.add_parser('show', help='show settings values')
    p_sshp = p_ssh.add_subparsers(title='sub_setting_param',
                                  dest='sub_setting_param', required=False)
    p_sshp.add_parser('project_search_dirs', help='show project search dirs')
    p_sshp.add_parser('tools_install_dir', help='show tools install dir')

    p_sse = p_ss.add_parser('set', help='set settings values')
    p_ssep = p_sse.add_subparsers(title='sub_setting_param', dest='sub_setting_param')
    p_ssep_t = p_ssep.add_parser('tools_install_dir', help='set tools install dir')
    p_ssep_t.add_argument('value',
                          help='path to the tools install dir').complete = shtab.DIRECTORY

    args = parser.parse_args()
    if 'func' in args.__dict__:
        args.func(args)

    if args.version:
        print(__title__ + "  v " + __version__)
        exit()

    if any(vars(args).values()):
        exit()

    if show_help_when_empty:
        parser.print_usage()

    return []



from __future__ import annotations  # for python 3.8

SETTINGS_FILENAME = "settings.json"


class SettingsData:
    run_when_os_starts: bool = None
    project_search_dirs: list[str] = None
    manually_installed_editors_search_dirs: list[str] = None
    tools_install_dir: str = None





from __future__ import annotations  # for python 3.8

from agstoolbox.core.ags.ags_editor import AgsEditor


class Release(AgsEditor):
    """a GitHub release object"""
    url: str = None
    id: str = None
    tag: str = None

    html_url: str = None

    is_pre_release: bool = False
    text_details: str = None
    published_at: str = None
    published_at_timestamp: float = None
    archive_name: str = None
    archive_url: str = None
    archive_size = None
    archive_id: str = None





from __future__ import annotations  # for python 3.8

import os
from pathlib import Path

from agstoolbox.core.settings.settings_data import SETTINGS_FILENAME, SettingsData
from agstoolbox.core.settings.settings_utils import load_settings_data_from_json_string, \
    save_settings_data_to_json_string, win_get_default_editor_search_dirs
from agstoolbox.core.utils.file import mkdirp, get_unique_valid_dirs
from agstoolbox.core.utils.singleton import Singleton
from agstoolbox import __title__
from agstoolbox.core.utils.startup import remove_app_at_startup, set_app_at_startup
from agstoolbox.core.utils.systemdirs import get_user_cache_dir, get_user_data_dir, \
    get_user_log_dir, get_user_documents_dir

app_name = __title__
app_author = "eri0o"


class StaticSettings:
    double_click_interval = 400
    DEFAULT_EXT_EDITORS_SEARCH_DIRS = win_get_default_editor_search_dirs()

    cache_dir = get_user_cache_dir(app_name, app_author)
    data_dir = get_user_data_dir(app_name, app_author)
    log_dir = get_user_log_dir(app_name, app_author)
    user_docs = get_user_documents_dir()

    DEFAULT_TOOLS_INSTALL_DIR = Path(os.path.join(user_docs, 'AgsToolbox')).as_posix()
    DEFAULT_PROJECTS_SEARCH_DIRS = [user_docs]

    DEFAULT_MAIN_PANEL_WIDTH = 320
    DEFAULT_MAIN_PANEL_HEIGHT = 512


class ConstSettings(StaticSettings, metaclass=Singleton):
    pass


def get_settings_dir():
    return Path(ConstSettings().data_dir).as_posix()


def get_settings_path():
    return Path(
        os.path.join(os.path.abspath(ConstSettings().data_dir), SETTINGS_FILENAME)).as_posix()


class BaseSettings:
    agstoolbox_package_install = ConstSettings.DEFAULT_TOOLS_INSTALL_DIR
    editor_base_install_dir = Path(os.path.join(agstoolbox_package_install, 'Editor')).as_posix()
    manually_installed_editors_search_dirs = ConstSettings().DEFAULT_EXT_EDITORS_SEARCH_DIRS

    project_search_dirs = ConstSettings().DEFAULT_PROJECTS_SEARCH_DIRS
    editor_install_dir = editor_base_install_dir
    tools_install_dir = agstoolbox_package_install
    run_when_os_starts: bool = False
    app_path: str = ""

    def set_app_path(self, path: str):
        self.app_path = path

    def set_manually_installed_editors_search_dirs(self, value: list[str]):
        if value is None:
            return

        self.manually_installed_editors_search_dirs = get_unique_valid_dirs(value)

    def get_manually_installed_editors_search_dirs(self) -> list[str]:
        if type(self.manually_installed_editors_search_dirs) == type(list()):
            return get_unique_valid_dirs(self.manually_installed_editors_search_dirs)

        return []

    def set_project_search_dirs(self, value: list[str]):
        if value is None:
            return

        self.project_search_dirs = get_unique_valid_dirs(value)

    def get_project_search_dirs(self) -> list[str]:
        if type(self.project_search_dirs) == type(list()):
            return get_unique_valid_dirs(self.project_search_dirs)

        return []

    def set_tools_install_dir(self, value):
        if value is None:
            return

        if value == "":
            value = ConstSettings.DEFAULT_TOOLS_INSTALL_DIR

        if not value == ConstSettings.DEFAULT_TOOLS_INSTALL_DIR:
            if not Path(value).exists():
                raise ValueError("Invalid path, doesn't exist")

        self.tools_install_dir = value
        self.editor_install_dir = Path(os.path.join(self.tools_install_dir, 'Editor')).as_posix()

    def set_run_when_os_starts(self, value: bool):
        if value is None or self.app_path is None or self.app_path == "":
            return

        if value is False:
            remove_app_at_startup(__title__)
            self.run_when_os_starts = False

        if value is True and self.app_path is not None and self.app_path != "":
            self.run_when_os_starts = set_app_at_startup(__title__, self.app_path)

    def get_tools_install_dir(self):
        return self.tools_install_dir

    def get_editor_install_dir(self):
        return self.editor_install_dir

    def get_run_when_os_starts(self):
        return self.run_when_os_starts

    def dump(self) -> str:
        data = SettingsData()
        data.run_when_os_starts = self.run_when_os_starts
        data.project_search_dirs = self.project_search_dirs
        data.manually_installed_editors_search_dirs = self.manually_installed_editors_search_dirs
        data.tools_install_dir = self.tools_install_dir

        return save_settings_data_to_json_string(data)

    def save(self):
        settings_dir = get_settings_dir()
        settings_path = get_settings_path()

        mkdirp(settings_dir)
        data_string: str = self.dump()

        with open(settings_path, 'w+', encoding="utf-8") as f:
            f.write(data_string)
            f.flush()

    def load(self):
        settings_path: str = get_settings_path()
        if not Path(settings_path).exists():
            return

        data = None
        sd = None
        with open(settings_path, 'r', encoding="utf-8") as f:
            data_string = f.read()
            if data_string is None:
                return
            sd = load_settings_data_from_json_string(data_string)

        self.set_manually_installed_editors_search_dirs(sd.manually_installed_editors_search_dirs)
        self.set_project_search_dirs(sd.project_search_dirs)
        self.set_tools_install_dir(sd.tools_install_dir)
        self.set_run_when_os_starts(sd.run_when_os_starts)

    def __init__(self):
        self.load()


class Settings(BaseSettings, metaclass=Singleton):
    pass



from PyQt6 import QtCore


class RunGuard:
    _shared: QtCore.QSharedMemory = None
    unique: bool = False

    def __init__(self, parent: QtCore.QObject, key: str):
        self._shared = QtCore.QSharedMemory(key, parent)
        self.unique = self._shared.create(512, QtCore.QSharedMemory.AccessMode.ReadWrite)

        if not self.unique:
            print('This app is already running!')



from __future__ import annotations  # for python 3.8
import os.path
from operator import attrgetter

import defusedxml.ElementTree as ETree

from agstoolbox.core.ags.game_project import GameProject, PROJECT_FILE_NAME
from agstoolbox.core.utils.file import get_dir, get_file_if_exists, get_gp_candidates_in_dir
from agstoolbox.core.version.version_utils import version_str_to_version

AGS_EDITOR_ROOT_TAG = 'AGSEditorDocument'


def text_file_starts_with_xml_Windows1252(filepath: str) -> bool:
    platform = ''
    try:
        with open(filepath, mode='r', encoding='cp1252') as file:
            platform = file.read(5)
    except FileNotFoundError:
        return False
    except OSError:
        return False
    finally:
        return platform == '<?xml'


def is_game_file(filepath: str) -> bool:
    if not os.path.exists(filepath):
        return False

    if not filepath.endswith(PROJECT_FILE_NAME):
        return False

    # it's too big, may crash parser later, better ignore for now
    if os.path.getsize(filepath) > 268435456:
        return False

    if not text_file_starts_with_xml_Windows1252(filepath):
        return False

    tree = ETree.parse(filepath)
    root = tree.getroot()
    if not root.tag == AGS_EDITOR_ROOT_TAG:
        return False

    if 'EditorVersion' not in root.attrib:
        return False

    return True


def gameagf_file_to_game_project(filepath: str) -> GameProject:
    gp = GameProject()
    tree = ETree.parse(filepath)
    root = tree.getroot()
    gp.path = filepath
    gp.directory = get_dir(filepath)
    ico_path = get_file_if_exists(gp.directory, "USER.ICO")
    if not ico_path:
        ico_path = get_file_if_exists(gp.directory, "template.ico")

    gp.ico_path = ico_path
    gp.last_modified = os.path.getmtime(filepath)

    gp.name = root.find('Game/Settings/GameName').text
    gp.ags_editor_version = version_str_to_version(root.attrib['EditorVersion'])
    gp.ags_editor_version_index = root.attrib['VersionIndex']
    return gp


def valid_gameagf_file_to_game_project(filepath: str) -> GameProject | None:
    if is_game_file(filepath):
        return gameagf_file_to_game_project(filepath)
    else:
        return None


def list_game_projects_in_dir(filepath: str) -> list[GameProject]:
    candidates = get_gp_candidates_in_dir(filepath, PROJECT_FILE_NAME)
    ags_projects = []

    for candidate in candidates:
        if not is_game_file(candidate):
            continue

        ags_projects.append(gameagf_file_to_game_project(candidate))

    unique = list(dict.fromkeys(ags_projects))
    ags_projects = unique
    ags_projects.sort(key=attrgetter("last_modified"), reverse=True)

    return ags_projects


def list_game_projects_in_dir_list(filepaths: list[str]) -> list[GameProject]:
    ags_projects = []
    for path in filepaths:
        candidates = list_game_projects_in_dir(path)
        ags_projects.extend(candidates)

    unique = list(dict.fromkeys(ags_projects))
    ags_projects = unique
    ags_projects.sort(key=attrgetter("last_modified"), reverse=True)

    return ags_projects


def get_unique_game_project_in_path(project_path: str) -> GameProject | None:
    game_project: GameProject | None = valid_gameagf_file_to_game_project(project_path)

    if game_project is None:
        projects: list[GameProject] = list_game_projects_in_dir(project_path)
        if len(projects) != 1:
            print('WARN: Invalid project path, not exactly 1 game project found!')
            return None

        game_project = projects[0]

    return game_project



import winreg
from pathlib import PureWindowsPath


def set_run_key(key, value):
    """
    Set/Remove Run Key in Windows registry.

    :param key: Run Key Name
    :param value: Program to Run
    :return: None
    """
    # This is for the system run variable
    reg_key = winreg.OpenKey(
        winreg.HKEY_CURRENT_USER,
        r'Software\Microsoft\Windows\CurrentVersion\Run',
        0, winreg.KEY_WRITE | winreg.KEY_READ)

    if value is not None:
        value = str(PureWindowsPath(value))

    with reg_key:
        if value is None:
            winreg.DeleteValue(reg_key, key)
        else:
            if '%' in value:
                var_type = winreg.REG_EXPAND_SZ
            else:
                var_type = winreg.REG_SZ
            winreg.SetValueEx(reg_key, key, 0, var_type, value)
        winreg.CloseKey(reg_key)


def get_run_key(key):
    reg_key = winreg.OpenKey(
        winreg.HKEY_CURRENT_USER,
        r'Software\Microsoft\Windows\CurrentVersion\Run',
        0, winreg.KEY_READ)

    with reg_key:
        ret_val = None
        try:
            ret_val = winreg.QueryValue(reg_key, key)
        except OSError:
            ret_val = None
        finally:
            winreg.CloseKey(reg_key)
            return ret_val
